/* Name: main.c
 * Project: 1-Key Keyboard
 * Author: Flip van den Berg - www.flipwork.nl
 * Creation Date: 2008-10-06
 * Based on AVR-USB drivers from Objective Developments - http://www.obdev.at/products/avrusb/index.html
 */

#include <avr/io.h>
#include <avr/wdt.h>
#include <avr/eeprom.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <util/delay.h>
#include <stdlib.h>

#include "usbdrv.h"
#include "oddebug.h"

#define BUTTON_PORT PORTB       /* PORTx - register for button output */
#define BUTTON_PIN PINB         /* PINx - register for button input */
#define BUTTON_BIT PB3          /* bit for button input/output */

/* ------------------------------------------------------------------------- */

static uchar    reportBuffer[3];    /* buffer for HID reports */
static uchar    idleRate;           /* in 4 ms units */
static uchar    reportCount;		/* current report */

static unsigned int        TimerDelay;                /* counter for delay period */
static uchar        sendInitialMove;	// Whether to override normal delay and random key in


/* ------------------------------------------------------------------------- */

PROGMEM char usbHidReportDescriptor[USB_CFG_HID_REPORT_DESCRIPTOR_LENGTH] = { /* USB report descriptor */
0x05, 0x01,                    // USAGE_PAGE (Generic Desktop)
0x09, 0x02,                    // USAGE (Mouse)
0xa1, 0x01,                    // COLLECTION (Application)
0x09, 0x01,                    //   USAGE (Pointer)
0xa1, 0x00,                    //   COLLECTION (Physical)
0x05, 0x09,                    //     USAGE_PAGE (Button)
0x19, 0x01,                    //     USAGE_MINIMUM (Button 1)
0x29, 0x03,                    //     USAGE_MAXIMUM (Button 3)
0x15, 0x00,                    //     LOGICAL_MINIMUM (0)
0x25, 0x01,                    //     LOGICAL_MAXIMUM (1)
0x95, 0x03,                    //     REPORT_COUNT (3)
0x75, 0x01,                    //     REPORT_SIZE (1)
0x81, 0x02,                    //     INPUT (Data,Var,Abs)
0x95, 0x01,                    //     REPORT_COUNT (1)
0x75, 0x05,                    //     REPORT_SIZE (5)
0x81, 0x03,                    //     INPUT (Cnst,Var,Abs)
0x05, 0x01,                    //     USAGE_PAGE (Generic Desktop)
0x09, 0x30,                    //     USAGE (X)
0x09, 0x31,                    //     USAGE (Y)
0x15, 0x81,                    //     LOGICAL_MINIMUM (-127)
0x25, 0x7f,                    //     LOGICAL_MAXIMUM (127)
0x75, 0x08,                    //     REPORT_SIZE (8)
0x95, 0x02,                    //     REPORT_COUNT (2)
0x81, 0x06,                    //     INPUT (Data,Var,Rel)
0xc0,                          //   END_COLLECTION
0xc0                           // END_COLLECTION
};
/* Basic mouse interface for 3 button mouse:
 //     Byte | D7      D6      D5      D4      D3      D2      D1      D0
 //    ------+---------------------------------------------------------------------
 //      0   |  0       0       0       0       0     Middle  Right   Left (Button)
 //      1   |                             X
 //      2   |                             Y
 //struct mouse_report_t
 //{
 //    uint8_t buttons;
 //    int8_t x;
 //    int8_t y;
 //}
 */

static uchar randMove(void)
{
	uchar RandNum;
	RandNum = rand();
	if (RandNum > 128) {RandNum |= 0xE0;} else {RandNum &= 0x1F;} // limit RandNum to -32 to + 31
	return RandNum;
}
static void buildReport(void)
{
uchar MoveX = 0; //if not changed by the if-statement below, then send an empty report
uchar MoveY = 0; //if not changed by the if-statement below, then send an empty report

    if(reportCount == 0){
		
		MoveX = randMove();
		MoveY = randMove();

		if( sendInitialMove > 0 )
		{
			switch (sendInitialMove) {
				case 1:
					MoveX = 0x14;	// +20
					MoveY = 0;
					break;
				case 2:
					MoveX = 0;
					MoveY = 0x14;	// +20
					break;
				case 3:
					MoveX = 0xEC;	// -20
					MoveY = 0;
					break;
				case 4:
					MoveX = 0;
					MoveY = 0xEC;	// -20
					break;
				default:
					MoveX = 0;
					MoveY = 0;
			}
			sendInitialMove--;
		}
    }

	reportCount++;									 //     Byte | D7      D6      D5      D4      D3      D2      D1      D0
    reportBuffer[0] = 0;		/* no buttons */	 //      0   |  0       0       0       0       0     Middle  Right   Left (Button)
    reportBuffer[1] = MoveX;	/* move X*/			 //      1   |                             X
    reportBuffer[2] = MoveY;	/* move Y*/			 //      2   |                             Y
}

static void timerPoll(void)
{
	static unsigned int timerCnt;

    if(TIFR & (1 << TOV1))
	{
        TIFR = (1 << TOV1); /* clear overflow */

		/* check for end of pseudorandom delay */
        if( ++timerCnt >= TimerDelay )	
		{
			//TimerDelay = 1260 + (rand() && 0x3FFF);		/* 1/63s * 63 * 20 + 0...32767 = 20sec to 4min40sec*/
			TimerDelay = 063 + (rand() && 0xFF);		/* 1/63s * 63  + 0...255 = 1sec to 5sec*/
			//TimerDelay = 63; // 1 sec
			//TimerDelay = 315; // DonP Test, 5 seconds constant 

			// DonP - ensure delay is overridden to ~1 second if this flag is set.
			// Actually only if it's greater than 1 otherwise the next event AFTER
			// the initial caps lock toggle is 1 second after the last toggle.
			// In other words if the flag is 1 the NEXT delay is what we want to be normal.
			// (Not the one after the flag is 0.)
			if( sendInitialMove > 1)
				TimerDelay = 15;

			timerCnt = 0;
			reportCount = 0; /* start report */
        }
    }
}

/* ------------------------------------------------------------------------- */

static void timerInit(void)
{
    TCCR1 = 0x0b;           /* select clock: 16.5M/1k -> overflow rate = 16.5M/256k = 62.94 Hz */
}

/* -------------------------------------------------------------------------------- */
/* ------------------------ interface to USB driver ------------------------ */
/* -------------------------------------------------------------------------------- */

uchar	usbFunctionSetup(uchar data[8])
{
usbRequest_t    *rq = (void *)data;

    usbMsgPtr = reportBuffer;
    if((rq->bmRequestType & USBRQ_TYPE_MASK) == USBRQ_TYPE_CLASS){    /* class request type */
        if(rq->bRequest == USBRQ_HID_GET_REPORT){  /* wValue: ReportType (highbyte), ReportID (lowbyte) */
            /* we only have one report type, so don't look at wValue */
            buildReport();
            return sizeof(reportBuffer);
        }else if(rq->bRequest == USBRQ_HID_GET_IDLE){
            usbMsgPtr = &idleRate;
            return 1;
        }else if(rq->bRequest == USBRQ_HID_SET_IDLE){
            idleRate = rq->wValue.bytes[1];
        }
    }else{
        /* no vendor specific requests implemented */
    }
	return 0;
}

/* ------------------------------------------------------------------------- */
/* ------------------------ Oscillator Calibration ------------------------- */
/* ------------------------------------------------------------------------- */

/* Calibrate the RC oscillator to 8.25 MHz. The core clock of 16.5 MHz is
 * derived from the 66 MHz peripheral clock by dividing. Our timing reference
 * is the Start Of Frame signal (a single SE0 bit) available immediately after
 * a USB RESET. We first do a binary search for the OSCCAL value and then
 * optimize this value with a neighboorhod search.
 * This algorithm may also be used to calibrate the RC oscillator directly to
 * 12 MHz (no PLL involved, can therefore be used on almost ALL AVRs), but this
 * is wide outside the spec for the OSCCAL value and the required precision for
 * the 12 MHz clock! Use the RC oscillator calibrated to 12 MHz for
 * experimental purposes only!
 */
static void calibrateOscillator(void)
{
uchar       step = 128;
uchar       trialValue = 0, optimumValue;
int         x, optimumDev, targetValue = (unsigned)(1499 * (double)F_CPU / 10.5e6 + 0.5);

    /* do a binary search: */
    do{
        OSCCAL = trialValue + step;
        x = usbMeasureFrameLength();    /* proportional to current real frequency */
        if(x < targetValue)             /* frequency still too low */
            trialValue += step;
        step >>= 1;
    }while(step > 0);
    /* We have a precision of +/- 1 for optimum OSCCAL here */
    /* now do a neighborhood search for optimum value */
    optimumValue = trialValue;
    optimumDev = x; /* this is certainly far away from optimum */
    for(OSCCAL = trialValue - 1; OSCCAL <= trialValue + 1; OSCCAL++){
        x = usbMeasureFrameLength() - targetValue;
        if(x < 0)
            x = -x;
        if(x < optimumDev){
            optimumDev = x;
            optimumValue = OSCCAL;
        }
    }
    OSCCAL = optimumValue;
}
/*
Note: This calibration algorithm may try OSCCAL values of up to 192 even if
the optimum value is far below 192. It may therefore exceed the allowed clock
frequency of the CPU in low voltage designs!
You may replace this search algorithm with any other algorithm you like if
you have additional constraints such as a maximum CPU clock.
For version 5.x RC oscillators (those with a split range of 2x128 steps, e.g.
ATTiny25, ATTiny45, ATTiny85), it may be useful to search for the optimum in
both regions.
*/

void    usbEventResetReady(void)
{
    calibrateOscillator();
    eeprom_write_byte(0, OSCCAL);   /* store the calibrated value in EEPROM */
}

/* ------------------------------------------------------------------------- */
/* --------------------------------- main ---------------------------------- */
/* ------------------------------------------------------------------------- */

int main(void)
{
uchar   i;
uchar   calibrationValue;
uint16_t resetCnt;

    calibrationValue = eeprom_read_byte(0); /* calibration value from last time */
    resetCnt = eeprom_read_word(1); /* reset counter */
	if (0 == ++resetCnt) {resetCnt=1;}
	eeprom_update_word(1,resetCnt);
    if(calibrationValue != 0xff){
        OSCCAL = calibrationValue;
    }
    
	//odDebugInit();
    usbInit();
    usbDeviceDisconnect();  /* enforce re-enumeration, do this while interrupts are disabled! */
    i = 0;
    while(--i){             /* fake USB disconnect for > 250 ms */
        wdt_reset();
        _delay_ms(1);
    }
    usbDeviceConnect();

    wdt_enable(WDTO_1S);

    timerInit();
	TimerDelay = 256; /* initial ~3 second delay */
	srand(resetCnt);
	sendInitialMove = 5; // DonP - set flag to override mouse moves this many times+1.
	// (+1 because buildreport runs at powerup once and we want to ignore that.)

    sei();

    for(;;){    /* main event loop */
        wdt_reset();
        usbPoll();

		/* A USB keypress cycle is defined as a scancode being present in a report, and
		then absent from a later report. To press and release the Caps Lock key, instead of
		holding it down, we need to send the report with the Caps Lock scancode and
		then an empty report. */
		
		if(usbInterruptIsReady() && reportCount < 2){ /* we can send another key */
        	buildReport();
           	usbSetInterrupt(reportBuffer, sizeof(reportBuffer));
        	}
		timerPoll();
	}
   	return 0;
}
